#
#  Expand variables
#

package Expander;
use Util qw(util_trim util_var diag_error diag_warning);
use strict;

############################### Bureaucracy ###############################

# Create new expander
sub new {
    my ($proto, $file) = @_;
    bless { file => $file,
            expandExpandable => 1,
            expandMake => 0,
            namespace => [],
            localVars => {},
            ignoredVars => {},
            usedVars => {}
          },
            ref($proto) || $proto;
}

# Clone this expander
sub clone {
    my $self = shift;
    { %$self }
}

# Add a local variable
sub addLocalVariable {
    my ($self, $name, $value) = @_;
    $self->{localVars}{$name} = $value;
}

# Set whether 'expandable' variables shall be expanded. This means
# variables marked as ->{expand} = 1, i.e. defined using ":=", which
# shall not appear in output.
sub setExpandExpandable {
    my $self = shift;
    my $doit = shift;
    $self->{expandExpandable} = $doit;
}

# Set whether Make variables shall be expanded. This means variables
# marked as ->{expand} = 0, i.e. defined using "=", which shall appear
# in output.
sub setExpandMake {
    my $self = shift;
    my $doit = shift;
    $self->{expandMake} = $doit;
}

# Add namespace. When looking for a variable X, first look for XNS,
# where NS is a namespace. This is used to implement context-dependant
# variable lookup.
sub addNamespace {
    my $self = shift;
    push @{$self->{namespace}}, @_;
}

# Get list of variables used during expansions using this Expander.
sub getUsedVariables {
    my $self = shift;
    keys %{$self->{usedVars}};
}

################################## Engine #################################

# Semi-Internal: This is the back-end to the expander functions.
# Splits a variable reference into list. The returned list will
# contain literals and variable names, alternatingly, starting with a
# literal. For example, "xx $(foo) $(bar)" will be split into ('xx ',
# 'foo', ' ', 'bar', '').
sub splitVars {
    my $self = shift;
    my $string = shift;
    my $file = $self->{file};
    my @result;

    pos($string) = 0;
    while ($string =~ m/\G([^\$]*)\$/gc) {
        if (@result & 1) {
            $result[-1] .= $1;
        } else {
            push @result, $1;
        }
        my $name = "";
        if ($string =~ m/\G\(/gc) {
            my $brace = 1;
            while ($brace) {
                if ($string =~ m/\G\(/gc) {
                    $name .= "("; ++$brace
                } elsif ($string =~ m/\G\)/gc) {
                    $name .= ")" if --$brace;
                } elsif ($string =~ m/\G([^()]+)/gc) {
                    $name .= $1;
                } else {
                    diag_error("malformed variable reference: '$string' unbalanced parens");
                }
            }
        } elsif ($string =~ m/\G(\*\*)/gc || $string =~ m/\G(.)/gc) {
            # "$**" is special Borland make syntax
            $name = $1
        } else {
            diag_error("malformed variable reference: '$string' ends in '\$'");
        }

        if ($name eq "\$") {
            $result[-1] .= '$$';
            next;
        }

        # if ($name =~ /^(\S+)\s+(.*)/) {
        #     # Function
        #     my $function = $1;
        #     my $args = split /,\s*/, $2;
        #     if ($function eq 'compile') {
        #     } elsif ($function eq 'ifeq') {
        #     } elsif ($function eq 'ifneq') {
        #     }
        # }

        my @cross = $self->expandCross($name);
        foreach (@cross) {
            my $name = $_;
            foreach (@{$self->{namespace}}) {
                my $newname = $name . $_;
                if (!exists($self->{ignoredVars}{$newname})
                    && (exists ($self->{localVars}{$newname})
                        || defined($file->getVar($newname))))
                {
                    my $newVar = $file->getVar($newname);
                    $self->{usedVars}{$newname} = 1;
                    if ($newVar->isAppend()) {
                        # We've resolved a variable reference "$(CFLAGS)" using
                        # "CFLAGS_foo += -O". Turn that into
                        # "$(CFLAGS) $(CFLAGS_foo)" to preserve append semantics.
                        $self->{usedVars}{$name} = 1;
                        push @result, $name, ' ';
                    }
                    $name = $newname;
                    last;
                }
            }
            $self->{usedVars}{$name} = 1;
            push @result, $name;
            push @result, ' ';
        }
        if (@cross) {
            pop @result;
        }
    }
    my $rest = substr($string,pos($string));
    if (@result & 1) {
        $result[-1] .= $rest
    } else {
        push @result, $rest;
    }

    # If last item is entirely blank, remove it
    if ($result[-1] !~ /\S/) { pop @result }

    @result;
}

# Expand variables
sub expandVars {
    my $self = shift;
    my $string = shift;
    my $result = "";
    my $literal = 1;
    foreach ($self->splitVars($string)) {
        if ($literal) {
            $result .= $_;
        } elsif (exists ($self->{ignoredVars}{$_})) {
            diag_warning ("Variable recursion on '$_'");
        } elsif (exists ($self->{localVars}{$_})) {
            $result .= $self->{localVars}{$_};
        } else {
            my $var = $self->{file}->getVar($_);
            if (defined($var)) {
                my $doit = 1;
                if ($var->isExpand()) {
                    if (!$self->{expandExpandable}) { $doit = 0 }
                } else {
                    if (!$self->{expandMake}) { $doit = 0 }
                }
                if ($doit) {
                    $self->{ignoredVars}{$_} = 1;
                    $result .= $self->expandVars($var->getStringValue());
                    delete $self->{ignoredVars}{$_};
                } else {
                    $result .= util_var($_);
                }
            } else {
                $result .= util_var($_);
            }
        }
        $literal ^= 1;
    }
    $result;
}

# Expand variables, cross-product semantics. If a variable yields
# several words, yield one output line per word. For example, if
#    a = 1 2
# expansion of
#    x$(a)y
# will yield x1y and x2y.
#
# This always expands all variables, as an actual Make utility does
# not have cross-product semantics.
sub expandCross {
    my $self = shift;
    my $string = shift;
    my @result = ("");
    my $literal = 1;
    foreach my $ele ($self->splitVars($string)) {
        if ($literal) {
            foreach (@result) { $_ .= $ele }
        } elsif (exists $self->{ignoredVars}{$ele}) {
            diag_warning ("Variable recursion on '$_'");
        } else {
            my $tmp;
            if (exists $self->{localVars}{$ele}) {
                $tmp = $self->{localVars}{$ele}
            } else {
                my $var = $self->{file}->getVar($ele);
                if (defined($var)) {
                    $self->{ignoredVars}{$ele} = 1;
                    $tmp = $self->expandVars($var->getStringValue());
                    delete $self->{ignoredVars}{$ele};
                }
            }
            if (defined($tmp)) {
                # Blank result, result will be null
                if ($tmp eq '') { return (); }
                # Build cross-product
                my @a = split /\s+/, $tmp;
                if (@a == 1) {
                    # Common case
                    foreach (@result) { $_ .= $a[0] }
                } else {
                    # General case
                    my @b = @result;
                    @result = ();
                    foreach my $b (@b) {
                        foreach my $a (@a) {
                            push @result, $b.$a;
                        }
                    }
                }
            } else {
                # Variable is not defined, so result will be null
                return ();
            }
        }
        $literal ^= 1;
    }
    @result;
}

# Expand variable references. Yields a string that contains the
# variable references expanded to our namespace/cross product rules,
# but does not expand the variables itself. For example,
#      $(CC) $(CFLAGS) -c $< -o $@
# might yield
#      $(CC) $(CFLAGS_foo.c) -c $< -o $@
# and
#      $(CC) .... $(LINK_$(DEPEND))
# might yield
#      $(CC)      $(LINK_one) $(LINK_two) $(LINK_three)
# However, it might well be possible that the result remains totally
# unchanged.
#
# Note that local variables are expanded in any case, as those never
# appear in output.
sub expandRefs {
    my $self = shift;
    my $string = shift;
    my $result = "";
    my $literal = 1;
    foreach ($self->splitVars($string)) {
        if ($literal) {
            $result .= $_;
        } else {
            if (exists $self->{localVars}{$_}) {
                $result .= $self->{localVars}{$_};
            } else {
                $result .= util_var($_);
            }
        }
        $literal ^= 1;
    }
    $result;
}

1;
